/*
 * Bluetooth Controlled 8x8 RGB LED Matrix
 *
 * Main Program
 *
 *    Bluetooth Pro Android App Button Assignments
 *    --------------------------------------------
 *      MoodLight     Random        Sequential
 *      Off           Speed         Off 
 *      Hue           Saturation    Value
 *      Message 1     Message 2     Message 3
 *       
 *
 * Written by: Craig A. Lindley
 * Last Update: 11/11/2014
 */

#include "Arduino.h"
#include "Modes.h"
#include "SerialEvent.h"
#include "Types.h"

// 12 button function definition strings
#define FN_MOODLIGHT  "mood@"
#define FN_RANDOM     "rand@"
#define FN_SEQUENTIAL  "seq@"
#define FN_OFF         "off@"
#define FN_SPEED      "speed"
#define FN_HUE          "hue"
#define FN_SAT         "sat@"
#define FN_VAL         "val@"
#define FN_MSG         "msg "

// Hue increment for 36 hue steps
#define HUE_INC 10

// Saturation increment for 10 steps
#define SAT_INC 0.1

// Value increment for 10 steps
#define VAL_INC 0.1

// Arbitrary maximum length of bluetooth command
#define MAX_BT_CMD_LENGTH 60

// Declare buffer sizes for serial event
#define TX_BUFFER_SIZE 16
#define RX_BUFFER_SIZE (MAX_BT_CMD_LENGTH)

// Constants
// States of the top level state machine
enum TLSMState {
    STATE_SETUP, STATE_CHECK_INPUT, STATE_PARSE, STATE_PREPARE, STATE_RUN
};

// Variable Declarations
enum TLSMState state;
boolean haveBTInput;
boolean updated;

// Declare a serial event instance for Teensy 3.1 serial port one
Serial1Event event;

// Create a global parameter instance
GLOBAL_PARAMETERS gParms;

// Declare mode instances
OffMode       offMode;
MoodLightMode moodLightMode(&gParms);
TextMsgMode   textMsgMode(&gParms);
PatternMode   patternMode(&gParms);

// Pointer to the current mode or pattern 
BaseClass *currentModeOrPattern;

// Buffer for blue tooth commnd
char btCmdBuffer[MAX_BT_CMD_LENGTH];

// Serial Event Bluetooth data received callback
void btDataReceived() {

    // Copy the bluetooth cmd into local buffer
    strcpy(btCmdBuffer, (char *) event.rxBuffer); 

    // Indicate there is Bluetooth data available
    haveBTInput = true;
}

// Run the top level state machine
void run() {

    char *ptr;

    switch (state) {
    case STATE_SETUP:
        {
            haveBTInput = false;
            updated = false;

            // Clear bluetooth buffer
            memset(btCmdBuffer, 0, sizeof(btCmdBuffer));

            // Sequential pattern mode is the default
            currentModeOrPattern = &patternMode;

            // Set next state
            state = STATE_CHECK_INPUT;
        }
        break;

    case STATE_CHECK_INPUT:
        {
            // Is there Bluetooth data available ?
            if (haveBTInput) {
                haveBTInput = false;
                state = STATE_PARSE;
            }    
            else    {
                state = STATE_RUN;
            }
        }
        break;

    case STATE_PARSE:
        {
            // Parse the newly received bluetooth data

            // Mood light mode ?
            if (strstr(btCmdBuffer, FN_MOODLIGHT) != NULL) {
                currentModeOrPattern = &moodLightMode;
                state = STATE_PREPARE;
            }    

            // Random Pattern mode ?
            else if (strstr(btCmdBuffer, FN_RANDOM) != NULL) {
                gParms.sequentialPatterns = false;
                currentModeOrPattern = &patternMode;
                state = STATE_PREPARE;
            }    

            // Sequential Pattern mode ?
            else if (strstr(btCmdBuffer, FN_SEQUENTIAL) != NULL) {
                gParms.sequentialPatterns = true;
                currentModeOrPattern = &patternMode;
                state = STATE_PREPARE;
            }    

            // Off mode ?
            else if (strstr(btCmdBuffer, FN_OFF) != NULL) {
                currentModeOrPattern = &offMode;
                state = STATE_PREPARE;
            }

            // Speed change ?
            else if ((ptr = strstr(btCmdBuffer, FN_SPEED)) != NULL) {
                // Move past the speed token
                ptr += 5;

                // Does @ immediately follow speed token ?
                if (*ptr == '@') {
                    // Yes so just increment the speed
                    gParms.speed++;
                    if (gParms.speed > 10) {
                        gParms.speed = 1;
                    }
                }    
                else    {
                    // No so a speed must have been specified
                    gParms.speed = atoi(ptr);

                    if (gParms.speed > 10) {
                        gParms.speed = 10;
                    }    
                    else if (gParms.speed < 1) {
                        gParms.speed = 1;
                    }
                }
                updated = true;
                state = STATE_RUN;
            }

            // Hue increment mode ?
            else if ((ptr = strstr(btCmdBuffer, FN_HUE)) != NULL) {
                // Move past the hue token
                ptr += 3;

                // Does @ immediately follow hue token ?
                if (*ptr == '@') {
                    // Yes so just increment the hue
                    gParms.hue += HUE_INC;
                    if (gParms.hue >= 360.0) {
                        gParms.hue = 0.0;
                    }
                }    
                else    {
                    // No so a hue must have been specified
                    gParms.hue = atof(ptr);
                    if (gParms.hue >= 360.0) {
                        gParms.hue = 359.9;
                    }    
                    else if (gParms.hue < 0) {
                        gParms.hue = 0.0;
                    }
                }
                updated = true;
                state = STATE_RUN;
            }

            // Sat increment mode ?
            else if (strstr(btCmdBuffer, FN_SAT) != NULL) {
                gParms.sat += SAT_INC;
                if (gParms.sat > 1.0) {
                    gParms.sat = 0.0;
                }
                updated = true;
                state = STATE_RUN;
            }

            // Val increment mode ?
            else if (strstr(btCmdBuffer, FN_VAL) != NULL) {
                gParms.val += VAL_INC;
                if (gParms.val > 1.0) {
                    gParms.val = 0.0;
                }
                updated = true;
                state = STATE_RUN;
            }

            // Text Msg ?
            else if (strstr(btCmdBuffer, FN_MSG) != NULL) {
                // Ptr to first char of msg
                char *ptr1 = strstr(btCmdBuffer, FN_MSG) + 4;

                // Ptr to termination char
                char *ptr2 = strchr(btCmdBuffer, '@');

                // Calculate msg length
                int len = ptr2 - ptr1;

                // Copy text msg to buffer
                strncpy(gParms.textMsgBuffer, ptr1, len);

                // Select text msg mode
                currentModeOrPattern = &textMsgMode;

                state = STATE_PREPARE;
            }
        }
        break;

    case STATE_PREPARE:
        {
            if (currentModeOrPattern != NULL) {
                // Prepare selected mode or pattern for running
                currentModeOrPattern->initStateMachine();

                // Set next state
                state = STATE_RUN;
            }    
            else    {
                // Set next state
                state = STATE_CHECK_INPUT;
            }            
        }
        break;

    case STATE_RUN:
        {
            if (currentModeOrPattern != NULL) {
                // Run the selected mode or pattern
                currentModeOrPattern->runStateMachine(updated);

                // Clear updated flag
                updated = false;
            }
            // Set the next state
            state = STATE_CHECK_INPUT;
        }
        break;
    }
    // Delay for a short period of time
    delay(10);
}

/*******************************************************************/
/***                      Main Program Setup                     ***/
/*******************************************************************/

void setup() {

    // Setup serial interfae
    Serial.begin(250000);

    // Wait for serial interface to settle
    delay(2000);

    // Initialize global parameters
    // First set everything to zero
    memset(&gParms, 0, sizeof(gParms));

    // Initialize non zero members
    gParms.hue = 240.0;
    gParms.sat = 1.0;
    gParms.val = 1.0;
    gParms.speed = 5;
    gParms.sequentialPatterns = true;

    // Initialize LED Matrix driver
    initializeDriver();

    // Initialize the SerialEvent for Bluetooth data reception

    // Declare transmit buffer size even though we never transmit from this code
    SERIAL1_MEMORY_TX(TX_BUFFER_SIZE);

    // Declare receive buffer size. Must be big enough for longest text message
    SERIAL1_MEMORY_RX(RX_BUFFER_SIZE);

    event.loopBack = false;                   // No internal loopback
    event.rxEventHandler = btDataReceived;    // Bluetooth data received callback
    event.rxTermCharacter = '@';              // Bluetooth messages terminated with @ char
    event.begin(9600);                        // Start serial port at Bluetooth HC-05 default rate

    // Set the initial state of the top level state machine
    state = STATE_SETUP;
}

/*******************************************************************/
/***                      Main Program Loop                      ***/
/*******************************************************************/

void loop() {

    // Continuously run the top level state machine
    run();
}



















